//+------------------------------------------------------------------+
//|                                              IFS-Fern_OpenCL.mq5 |
//|                        Copyright 2011, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, MetaQuotes Software Corp."
#property link      "http://www.mql5.com"
#property version   "1.00"

#define SIZE_X 600
#define SIZE_Y 600

//-- Barnsley's fern
float IFS[]=
  {
   0.00f,  0.85f,  0.20f,  -0.15f,
   0.00f,  0.04f, -0.26f,   0.28f,
   0.00f, -0.04f,  0.23f,   0.26f,
   0.16f,  0.85f,  0.22f,   0.24f,
   0.00f,  0.00f,  0.00f,   0.00f,
   0.00f,  1.60f,  1.60f,   0.00f,
   10,860,930,1000
  };
//+------------------------------------------------------------------+
//|   OpenCL                                               |
//+------------------------------------------------------------------+
const string cl_src=
                    "__kernel void IFS_Fern(                                                       \r\n"
                    "                       float xx0,                                             \r\n"
                    "                       float yy0,                                             \r\n"
                    "                       int xsize,                                             \r\n"
                    "                       int ysize,                                             \r\n"
                    "                       float scale,                                           \r\n"
                    "                       float maxincr,                                         \r\n"
                    "                       float xc,                                              \r\n"
                    "                       float yc,                                              \r\n"
                    "                       __global float *IFS,                                   \r\n"
                    "                       __global uint *vs)                                     \r\n"
                    "             {                                                                \r\n"
                    "                float x0=xx0;                                                 \r\n"
                    "                float y0=yy0;                                                 \r\n"
                    "                float x;                                                      \r\n"
                    "                float y;                                                      \r\n"
                    "                float xcenter=xc;                                             \r\n"
                    "                float ycenter=yc;                                             \r\n"
                    "                float iter=0;                                                 \r\n"
                    "                float realincrements=0;                                       \r\n"
                    "                uint next=get_global_id(0)*13;                                \r\n"
                    "                do                                                            \r\n"
                    "                  {                                                           \r\n"
                    "                   next=next*1103515245+12345;                                \r\n"
                    "                   float prb=1000*((next/65536)%32768)/32767.0;               \r\n"
                    "                   for(int k=0; k<4; k++)                                     \r\n"
                    "                     {                                                        \r\n"
                    "                      if(prb<=IFS[24+k])                                      \r\n"
                    "                        {                                                     \r\n"
                    "                         x = IFS[k] * x0 + IFS[4+k] * y0 + IFS[16+k];         \r\n"
                    "                         y = IFS[8+k] * x0 + IFS[12+k] * y0 + IFS[20+k];      \r\n"
                    "                         int scX = (int)rint(xsize/2 + (x-xcenter)*scale);    \r\n"
                    "                         int scY = (int)rint(ysize/2 + (y-ycenter)*scale);    \r\n"
                    "                        if(iter>10)                                           \r\n"
                    "                          { if(scX>=0 && scX<xsize && scY>=0 && scY<ysize)    \r\n"
                    "                           {                                                  \r\n"
                    "                            vs[xsize*scY+scX]+=1;                             \r\n"
                    "                            realincrements=realincrements+1;                  \r\n"
                    "                           }                                                  \r\n"
                    "                          }                                                   \r\n"
                    "                         break;                                               \r\n"
                    "                        }                                                     \r\n"
                    "                      x0 = x;                                                 \r\n"
                    "                      y0 = y;                                                 \r\n"
                    "                     }                                                        \r\n"
                    "                     iter=iter+1;                                             \r\n"
                    "                  }                                                           \r\n"
                    "                while(realincrements<maxincr);                                \r\n"
                    "               }                                                              \r\n";
//+------------------------------------------------------------------+
//|  BMP                                               |
//+------------------------------------------------------------------+
struct BitmapHeader
  {
   ushort            type;
   uint              size;
   uint              reserv;
   uint              offbits;
   uint              imgSSize;
   uint              imgWidth;
   uint              imgHeight;
   ushort            imgPlanes;
   ushort            imgBitCount;
   uint              imgCompression;
   uint              imgSizeImage;
   uint              imgXPelsPerMeter;
   uint              imgYPelsPerMeter;
   uint              imgClrUsed;
   uint              imgClrImportant;
  };
//+------------------------------------------------------------------+
//|                                                |
//+------------------------------------------------------------------+
bool SaveBitmapToFile(const string filename,uint &bitmap[],const BitmapHeader &info)
  {
//---  
   int file=FileOpen(filename,FILE_WRITE|FILE_BIN);
   if(file==INVALID_HANDLE)
     {
      Print(__FUNCTION__," error opening '",filename,"'");
      return(false);
     }
//---     
   if(FileWriteStruct(file,info)==sizeof(info))
     {
      if(FileWriteArray(file,bitmap)==ArraySize(bitmap))
        {
         FileClose(file);
         return(true);
        }
     }
//---  ,     
   FileClose(file);
   FileDelete(filename);
   Print(__FUNCTION__," error writting '",filename,"'");
//---  
   return(false);
  }

uchar Palette[768];
//+------------------------------------------------------------------+
//|                                        |
//+------------------------------------------------------------------+
void Blend(uchar c1,uchar c2,uchar r1,uchar g1,uchar b1,uchar r2,uchar g2,uchar b2)
  {
   int n=(c2-c1);
   for(int i=0; i<=n; i++)
     {
      if((c1+i+2)<ArraySize(Palette))
        {
         Palette[3*(c1+i)]=uchar(MathRound(1*(r1*(n-i)+r2*i)*1.0/n));
         Palette[3*(c1+i)+1]=uchar(MathRound(1*(g1*(n-i) + g2*i) * 1.0/ n));
         Palette[3*(c1+i)+2]=uchar(MathRound(1*(b1*(n-i) + b2*i) * 1.0/ n));
        }
     }
  }
//+------------------------------------------------------------------+
//| GetPalColor                                                      |
//+------------------------------------------------------------------+
int GetPalColor(uchar index)
  {
   int ind=index;
   if(ind<=0) {ind=0;}
   uchar r=Palette[3*(ind)];
   uchar g=Palette[3*(ind)+1];
   uchar b=Palette[3*(ind)+2];
//--- usually return(r+256*g+65536*b);
//--- but R and B must rotated in BMP!
   return(b+256*g+65536*r);
  }
//+------------------------------------------------------------------+
//| VirtualScreenToBMP                                               |
//+------------------------------------------------------------------+
void VirtualScreenToBMP(int xsize,int ysize,uint &vs[])
  {
   for(int i=0; i<xsize; i++)
     {
      for(int j=0; j<ysize; j++)
        {
         int pos=int(xsize*j+i);
         uint colind=vs[pos];
         uint xcol=GetPalColor(uchar(colind));
         if(colind==0) xcol=0x00;
         vs[pos]=xcol;
        }
     }
  }
int cl_ctx;
int cl_prg;
int cl_krn;
int cl_mem,cl_mem2;

float centerx=0;
float centery=5;

float scale=50;
//+------------------------------------------------------------------+
//| ShowImage                                                        |
//+------------------------------------------------------------------+
void ShowImage()
  {
//---   
   uint  offset[]={0};
   uint  work  []={500};
//---     OpenCL
   CLSetKernelArgMem(cl_krn,8,cl_mem2);
   CLSetKernelArgMem(cl_krn,9,cl_mem);
//---     
   uint buf[];
   ArrayResize(buf,SIZE_X*SIZE_Y);
//---  
   BitmapHeader info;
   ZeroMemory(info);
   info.type          =0x4d42;
   info.size          =sizeof(info)+SIZE_X*SIZE_Y*4;
   info.offbits       =sizeof(info);
   info.imgSSize      =40;
   info.imgWidth      =SIZE_X;
   info.imgHeight     =SIZE_Y;
   info.imgPlanes     =1;
   info.imgBitCount   =32;
   info.imgCompression=0;                // BI_RGB
   info.imgSizeImage  =SIZE_X*SIZE_Y*4;

//---     
   ObjectCreate(0,"Fern",OBJ_BITMAP_LABEL,0,0,0);
   ObjectSetInteger(0,"Fern",OBJPROP_XDISTANCE,4);
   ObjectSetInteger(0,"Fern",OBJPROP_YDISTANCE,25);

   ZeroMemory(buf);
//---    OpenCL
   CLSetKernelArg(cl_krn,2,SIZE_X);
   CLSetKernelArg(cl_krn,3,SIZE_Y);

//---  
   float xc=centerx;
   float yc=centery;

   CLSetKernelArg(cl_krn,6,xc);
   CLSetKernelArg(cl_krn,7,yc);

   ZeroMemory(buf);
   CLBufferWrite(cl_mem,buf);

   uint x=GetTickCount();
   float pointsneeded=float(MathRound(1500+scale));

//---        OpenCL
   float xx0=float(MathRand()/32757.0);
   float yy0=float(MathRand()/32757.0);

//---   
   CLSetKernelArg(cl_krn,4,scale);
   CLSetKernelArg(cl_krn,5,pointsneeded);

   CLSetKernelArg(cl_krn,0,xx0);
   CLSetKernelArg(cl_krn,1,yy0);

//---  
   CLExecute(cl_krn,1,offset,work);
   CLBufferRead(cl_mem,buf);
//---    bmp
   VirtualScreenToBMP(SIZE_X,SIZE_Y,buf);

//---   
   Comment(IntegerToString(GetTickCount()-x)+" msec per frame ");
//---       
   string s="Fern"+IntegerToString(1,4,'0')+".bmp";
   SaveBitmapToFile(s,buf,info);
   ObjectSetString(0,"Fern",OBJPROP_BMPFILE,NULL);
   ObjectSetString(0,"Fern",OBJPROP_BMPFILE,"\\Files\\"+s);
   ChartRedraw();
  }
//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
int OnInit()
  {
//---grey by default
   Blend(0,255,0,0,0,255,255,255);
   Blend(1,8,0,25,0,10,52,10);
   Blend(8,43,10,52,10,200,147,8);
   Blend(44,127,200,147,8,255,255,255);

//---  OpenCL 
   if((cl_ctx=CLContextCreate())==INVALID_HANDLE)
     {
      Print("OpenCL not found");
      return(1);
     }
   if((cl_prg=CLProgramCreate(cl_ctx,cl_src))==INVALID_HANDLE)
     {
      CLContextFree(cl_ctx);
      Print("OpenCL program create failed");
      return(1);
     }
   if((cl_krn=CLKernelCreate(cl_prg,"IFS_Fern"))==INVALID_HANDLE)
     {
      CLProgramFree(cl_prg);
      CLContextFree(cl_ctx);
      Print("OpenCL kernel create failed");
      return(1);
     }
   if((cl_mem=CLBufferCreate(cl_ctx,SIZE_X*SIZE_Y*sizeof(uint),CL_MEM_READ_WRITE))==INVALID_HANDLE)
     {
      CLKernelFree(cl_krn);
      CLProgramFree(cl_prg);
      CLContextFree(cl_ctx);
      Print("OpenCL buffer create failed");
      return(1);
     }
   if((cl_mem2=CLBufferCreate(cl_ctx,28*sizeof(float),CL_MEM_READ_WRITE))==INVALID_HANDLE)
     {
      CLKernelFree(cl_krn);
      CLProgramFree(cl_prg);
      CLContextFree(cl_ctx);
      Print("OpenCL buffer create failed");
      return(1);
     }

   CLBufferWrite(cl_mem2,IFS);
   ShowImage();
   return(0);
  }
//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |
//+------------------------------------------------------------------+
void OnDeinit(const int r)
  {
   ObjectDelete(0,"Fern");
//---   OpenCL
   CLBufferFree(cl_mem);
   CLKernelFree(cl_krn);
   CLProgramFree(cl_prg);
   CLContextFree(cl_ctx);
  }
//+------------------------------------------------------------------+
//| ChartEvent function                                              |
//+------------------------------------------------------------------+
void OnChartEvent(const int id,         // Event identifier  
                  const long& lparam,   // Event parameter of long type
                  const double& dparam, // Event parameter of double type
                  const string& sparam  // Event parameter of string type
                  )
  {
//--- the mouse has been clicked on the graphic object
   if(id==CHARTEVENT_OBJECT_CLICK)
     {
      if(sparam=="Fern")
        {
         float x0 = centerx - (SIZE_X/2)/scale;
         float y0 = centery - (SIZE_Y/2)/scale;

         float x1 = centerx + (SIZE_X/2)/scale;
         float y1 = centery + (SIZE_Y/2)/scale;

         centerx = float (x0 + (x1-x0)*(1.0*lparam/SIZE_X));
         centery = float (y1 - (y1-y0)*(1.0*dparam/SIZE_Y));

         scale=float(scale*1.1);

         Print("X=",centerx," Y=",centery," scale=",scale);
         ShowImage();
        }
     }
  }
//+------------------------------------------------------------------+
